@@ -25,16 +25,28 @@
  *
  * @author Nicolai Ehemann <en@enlightened.de>
  * @author André Rothe <arothe@zks.uni-leipzig.de>
- * @copyright Copyright (C) 2013-2014 Nicolai Ehemann and contributors
+ * @copyright Copyright (C) 2013-2015 Nicolai Ehemann and contributors
  * @license GNU GPL
- * @version 0.4
+ * @version 0.7
  */
 namespace ZipStreamer;
 
 require "lib/Count64.php";
 
+class COMPR {
+  // compression method
+  const STORE = 0x0000; //  0 - The file is stored (no compression)
+  const DEFLATE = 0x0008; //  8 - The file is deflated
+
+  // compression level (for deflate compression)
+  const NONE = 0;
+  const NORMAL = 1;
+  const MAXIMUM = 2;
+  const SUPERFAST = 3;
+}
+
 class ZipStreamer {
-  const VERSION = "0.5";
+  const VERSION = "0.7";
 
   const ZIP_LOCAL_FILE_HEADER = 0x04034b50; // local file header signature
   const ZIP_CENTRAL_FILE_HEADER = 0x02014b50; // central file header signature
@@ -42,17 +54,22 @@ class ZipStreamer {
   const ZIP64_END_OF_CENTRAL_DIRECTORY = 0x06064b50; //zip64 end of central directory record
   const ZIP64_END_OF_CENTRAL_DIR_LOCATOR = 0x07064b50; // zip64 end of central directory locator
 
-  //TODO: make this dynamic, depending on flags/compression methods
-  const ATTR_VERSION_TO_EXTRACT = 0x2d; // version needed to extract (min. 4.5)
   const ATTR_MADE_BY_VERSION = 0x032d; // made by version  (upper byte: UNIX, lower byte v4.5)
 
-  const STREAM_CHUNK_SIZE = 1048576; // 1mb chunks
+  const STREAM_CHUNK_SIZE = 1048560; // 16 * 65535 = almost 1mb chunks, for best deflate performance
 
   private $extFileAttrFile;
   private $extFileAttrDir;
 
   /** @var stream output stream zip file is written to */
   private $outStream;
+  /** @var boolean zip64 enabled */
+  private $zip64 = True;
+  /** @var int compression method */
+  private $compress;
+  /** @var int compression level */
+  private $level;
+
   /** @var array central directory record */
   private $cdRec = array();
   /** @var int offset of next file to be added */
@@ -61,11 +78,23 @@ class ZipStreamer {
   private $isFinalized = false;
 
   /**
-   * Constructor.
+   * Constructor. Initializes ZipStreamer object for immediate usage.
+   * @param array $options Optional, ZipStreamer and zip file options as key/value pairs.
+   *                       Valid options are:
+   *                       * outstream: stream the zip file is output to (default: stdout)
+   *                       * zip64: enabled/disable zip64 support (default: True)
+   *                       * compress: int, compression method (one of COMPR::STORE,
+   *                                   COMPR::DEFLATE, default COMPR::STORE)
+   *                                   can be overridden for single files
+   *                       * level: int, compression level (one of COMPR::NORMAL,
+   *                                COMPR::MAXIMUM, COMPR::SUPERFAST, default COMPR::NORMAL)
    */
   function __construct($options = NULL) {
     $defaultOptions = array(
         'outstream' => NULL,
+        'zip64' => True,
+        'compress' => COMPR::STORE,
+        'level' => COMPR::NORMAL,
     );
     if (is_null($options)) {
       $options = array();
@@ -77,19 +106,23 @@ class ZipStreamer {
     } else {
       $this->outstream = fopen('php://output', 'w');
     }
+    $this->zip64 = $options['zip64'];
+    $this->compress = $options['compress'];
+    $this->level = $options['level'];
+    $this->validateCompressionOptions($this->compress, $this->level);
     //TODO: is this advisable/necessary?
     if (ini_get('zlib.output_compression')) {
       ini_set('zlib.output_compression', 'Off');
     }
     // initialize default external file attributes
     $this->extFileAttrFile = UNIX::getExtFileAttr(UNIX::S_IFREG |
-                                                  UNIX::S_IRUSR | UNIX::S_IXUSR | UNIX::S_IRGRP |
+                                                  UNIX::S_IRUSR | UNIX::S_IWUSR | UNIX::S_IRGRP |
                                                   UNIX::S_IROTH);
     $this->extFileAttrDir = UNIX::getExtFileAttr(UNIX::S_IFDIR |
                                                  UNIX::S_IRWXU | UNIX::S_IRGRP | UNIX::S_IXGRP |
                                                  UNIX::S_IROTH | UNIX::S_IXOTH) |
                             DOS::getExtFileAttr(DOS::DIR);
-    $this->offset = Count64::construct(0);
+    $this->offset = Count64::construct(0, !$this->zip64);
   }
 
   function __destruct() {
@@ -97,6 +130,17 @@ class ZipStreamer {
     $this->cdRec = null;
   }
 
+  private function getVersionToExtract($isDir) {
+    if ($this->zip64) {
+      $version = 0x2d; // 4.5 - File uses ZIP64 format extensions
+    } else if ($isDir) {
+      $version = 0x14; // 2.0 - File is a folder (directory)
+    } else {
+      $version = 0x0a; //   1.0 - Default value
+    }
+    return $version;
+  }
+
   /**
   * Send appropriate http headers before streaming the zip file and disable output buffering.
   * This method, if used, has to be called before adding anything to the zip file.
@@ -136,15 +180,29 @@ class ZipStreamer {
    *
    * @param string $stream      Stream to read data from
    * @param string $filePath    Filepath and name to be used in the archive.
-   * @param int    $timestamp   (Optional) Timestamp for the added file, if omitted or set to 0, the current time will be used.
-   * @param string $fileComment (Optional) Comment to be added to the archive for this file. To use fileComment, timestamp must be given.
-   * @param bool   $compress    (Optional) Compress file, if set to false the file will only be stored. Default FALSE.
+   * @param array $options      Optional, additional options
+   *                            Valid options are:
+   *                               * int timestamp: timestamp for the file (default: current time)
+   *                               * string comment: comment to be added for this file (default: none)
+   *                               * int compress: compression method (override global option for this file)
+   *                               * int level: compression level (override global option for this file)
    * @return bool $success
    */
-  public function addFileFromStream($stream, $filePath, $timestamp = 0, $fileComment = null, $compress = false) {
+  public function addFileFromStream($stream, $filePath, $options = NULL) {
     if ($this->isFinalized) {
       return false;
     }
+    $defaultOptions = array(
+        'timestamp' => NULL,
+        'comment' => NULL,
+        'compress' => $this->compress,
+        'level' => $this->level,
+    );
+    if (is_null($options)) {
+    	$options = array();
+    }
+    $options = array_merge($defaultOptions, $options);
+    $this->validateCompressionOptions($options['compress'], $options['level']);
 
     if (!is_resource($stream) || get_resource_type($stream) != 'stream') {
       return false;
@@ -153,21 +211,18 @@ class ZipStreamer {
     $filePath = self::normalizeFilePath($filePath);
 
     $gpFlags = GPFLAGS::ADD;
-    if ($compress) {
-      $gzMethod = GZMETHOD::DEFLATE;
-    } else {
-      $gzMethod = GZMETHOD::STORE;
-    }
 
-    list($gpFlags, $lfhLength) = $this->beginFile($filePath, $fileComment, $timestamp, $gpFlags, $gzMethod);
-    list($dataLength, $gzLength, $dataCRC32) = $this->streamFileData($stream, $compress);
+    list($gpFlags, $lfhLength) = $this->beginFile($filePath, False, $options['comment'], $options['timestamp'], $gpFlags, $options['compress']);
+    list($dataLength, $gzLength, $dataCRC32) = $this->streamFileData($stream, $options['compress'], $options['level']);
+
+    $ddLength = $this->addDataDescriptor($dataLength, $gzLength, $dataCRC32);
 
     // build cdRec
-    $this->cdRec[] = $this->buildCentralDirectoryHeader($filePath, $timestamp, $gpFlags, $gzMethod,
-        $dataLength, $gzLength, $dataCRC32, $this->extFileAttrFile);
+    $this->cdRec[] = $this->buildCentralDirectoryHeader($filePath, $options['timestamp'], $gpFlags, $options['compress'],
+        $dataLength, $gzLength, $dataCRC32, $this->extFileAttrFile, False);
 
     // calc offset
-    $this->offset->add($lfhLength)->add($gzLength);
+    $this->offset->add($ddLength)->add($lfhLength)->add($gzLength);
 
     return true;
   }
@@ -176,26 +231,35 @@ class ZipStreamer {
    * Add an empty directory entry to the zip archive.
    *
    * @param string $directoryPath  Directory Path and name to be added to the archive.
-   * @param int    $timestamp      (Optional) Timestamp for the added directory, if omitted or set to 0, the current time will be used.
-   * @param string $fileComment    (Optional) Comment to be added to the archive for this directory. To use fileComment, timestamp must be given.
+   * @param array $options      Optional, additional options
+   *                            Valid options are:
+   *                               * int timestamp: timestamp for the file (default: current time)
+   *                               * string comment: comment to be added for this file (default: none)
    * @return bool $success
    */
-  public function addEmptyDir($directoryPath, $timestamp = 0, $fileComment = null) {
+  public function addEmptyDir($directoryPath, $options = NULL) {
     if ($this->isFinalized) {
       return false;
     }
+    $defaultOptions = array(
+    		'timestamp' => NULL,
+    		'comment' => NULL,
+    );
+    if (is_null($options)) {
+    	$options = array();
+    }
+    $options = array_merge($defaultOptions, $options);
 
     $directoryPath = self::normalizeFilePath($directoryPath) . '/';
 
     if (strlen($directoryPath) > 0) {
-      $gpFlags = 0x0000; // Compression type 0 = stored
-      $gzMethod = GZMETHOD::STORE; // Compression type 0 = stored
-
-      list($gpFlags, $lfhLength) = $this->beginFile($directoryPath, $fileComment, $timestamp, $gpFlags, $gzMethod);
+      $gpFlags = 0x0000;
+      $gzMethod = COMPR::STORE; // Compression type 0 = stored
 
+      list($gpFlags, $lfhLength) = $this->beginFile($directoryPath, True, $options['comment'], $options['timestamp'], $gpFlags, $gzMethod);
       // build cdRec
-      $this->cdRec[] = $this->buildCentralDirectoryHeader($directoryPath, $timestamp, $gpFlags, $gzMethod,
-          0, 0, 0, $this->extFileAttrDir);
+      $this->cdRec[] = $this->buildCentralDirectoryHeader($directoryPath, $options['timestamp'], $gpFlags, $gzMethod,
+          Count64::construct(0, !$this->zip64), Count64::construct(0, !$this->zip64), 0, $this->extFileAttrDir, True);
 
       // calc offset
       $this->offset->add($lfhLength);
@@ -207,7 +271,8 @@ class ZipStreamer {
 
   /**
    * Close the archive.
-   * A closed archive can no longer have new files added to it.
+   * A closed archive can no longer have new files added to it. After
+   * closing, the zip file is completely written to the output stream.
    * @return bool $success
    */
   public function finalize() {
@@ -217,14 +282,16 @@ class ZipStreamer {
       $cd = implode('', $this->cdRec);
       $this->write($cd);
 
-      // print the zip64 end of central directory record
-      $this->write($this->buildZip64EndOfCentralDirectoryRecord(strlen($cd)));
+      if ($this->zip64) {
+        // print the zip64 end of central directory record
+        $this->write($this->buildZip64EndOfCentralDirectoryRecord(strlen($cd)));
 
-      // print the zip64 end of central directory locator
-      $this->write($this->buildZip64EndOfCentralDirectoryLocator(strlen($cd)));
+        // print the zip64 end of central directory locator
+        $this->write($this->buildZip64EndOfCentralDirectoryLocator(strlen($cd)));
+      }
 
       // print end of central directory record
-      $this->write($this->buildEndOfCentralDirectoryRecord());
+      $this->write($this->buildEndOfCentralDirectoryRecord(strlen($cd)));
 
       $this->flush();
 
@@ -237,6 +304,26 @@ class ZipStreamer {
     return false;
   }
 
+  private function validateCompressionOptions($compress, $level) {
+    if (COMPR::STORE === $compress) {
+    } else if (COMPR::DEFLATE === $compress) {
+      if (COMPR::NONE !== $level
+        && !class_exists(DeflatePeclStream::PECL1_DEFLATE_STREAM_CLASS)
+        && !class_exists(DeflatePeclStream::PECL2_DEFLATE_STREAM_CLASS)) {
+        throw new \Exception('unable to use compression method DEFLATE with level other than NONE (requires pecl_http >= 0.10)');
+      }
+    } else {
+      throw new \Exception('invalid option ' . $compress . ' (compression method)');
+    }
+
+    if (!(COMPR::NONE === $level ||
+        COMPR::NORMAL === $level ||
+        COMPR::MAXIMUM === $level ||
+        COMPR::SUPERFAST === $level)) {
+      throw new \Exception('invalid option ' . $level . ' (compression level');
+    }
+  }
+
   private function write($data) {
     return fwrite($this->outstream, $data);
   }
@@ -245,7 +332,7 @@ class ZipStreamer {
     return fflush($this->outstream);
   }
 
-  private function beginFile($filePath, $fileComment, $timestamp, $gpFlags = 0x0000, $gzMethod = GZMETHOD::STORE,
+  private function beginFile($filePath, $isDir, $fileComment, $timestamp, $gpFlags, $gzMethod,
       $dataLength = 0, $gzLength = 0, $dataCRC32 = 0) {
 
     $isFileUTF8 = mb_check_encoding($filePath, 'UTF-8') && !mb_check_encoding($filePath, 'ASCII');
@@ -257,31 +344,40 @@ class ZipStreamer {
     }
 
     $localFileHeader = $this->buildLocalFileHeader($filePath, $timestamp, $gpFlags, $gzMethod, $dataLength,
-        $gzLength, $dataCRC32);
+        $gzLength, $isDir, $dataCRC32);
 
     $this->write($localFileHeader);
 
     return array($gpFlags, strlen($localFileHeader));
   }
 
-  private function streamFileData($stream, $compress) {
-    $dataLength = Count64::construct(0);
-    $gzLength = Count64::construct(0);
+  private function streamFileData($stream, $compress, $level) {
+    $dataLength = Count64::construct(0, !$this->zip64);
+    $gzLength = Count64::construct(0, !$this->zip64);
     $hashCtx = hash_init('crc32b');
+    if (COMPR::DEFLATE === $compress) {
+      $compStream = DeflateStream::create($level);
+    }
 
     while (!feof($stream)) {
       $data = fread($stream, self::STREAM_CHUNK_SIZE);
       $dataLength->add(strlen($data));
       hash_update($hashCtx, $data);
-      if ($compress) {
-        //TODO: this is broken.
-        $data = gzdeflate($data);
+      if (COMPR::DEFLATE === $compress) {
+        $data = $compStream->update($data);
       }
       $gzLength->add(strlen($data));
       $this->write($data);
 
       $this->flush();
     }
+    if (COMPR::DEFLATE === $compress) {
+      $data = $compStream->finish();
+      $gzLength->add(strlen($data));
+      $this->write($data);
+
+      $this->flush();
+    }
     $crc = unpack('N', hash_final($hashCtx, true));
     return array($dataLength, $gzLength, $crc[1]);
   }
@@ -296,28 +392,55 @@ class ZipStreamer {
       . pack32le(0);             // number of the disk on which this file starts   4 bytes
   }
 
-  private function buildLocalFileHeader($filePath, $timestamp, $gpFlags = 0x0000,
-      $gzMethod = GZMETHOD::STORE, $dataLength, $gzLength, $dataCRC32 = 0) {
+  private function buildLocalFileHeader($filePath, $timestamp, $gpFlags,
+      $gzMethod, $dataLength, $gzLength, $isDir = False, $dataCRC32 = 0) {
+    $versionToExtract = $this->getVersionToExtract($isDir);
     $dosTime = self::getDosTime($timestamp);
-    $zip64Ext = $this->buildZip64ExtendedInformationField($dataLength, $gzLength);
+    if ($this->zip64) {
+      $zip64Ext = $this->buildZip64ExtendedInformationField($dataLength, $gzLength);
+      $dataLength = -1;
+      $gzLength = -1;
+    } else {
+      $zip64Ext = '';
+    }
 
     return ''
       . pack32le(self::ZIP_LOCAL_FILE_HEADER)   // local file header signature     4 bytes  (0x04034b50)
-      . pack16le(self::ATTR_VERSION_TO_EXTRACT) // version needed to extract       2 bytes
+      . pack16le($versionToExtract)             // version needed to extract       2 bytes
       . pack16le($gpFlags)                      // general purpose bit flag        2 bytes
       . pack16le($gzMethod)                     // compression method              2 bytes
       . pack32le($dosTime)                      // last mod file time              2 bytes
                                                 // last mod file date              2 bytes
       . pack32le($dataCRC32)                    // crc-32                          4 bytes
-      . pack32le(-1)                            // compressed size                 4 bytes
-      . pack32le(-1)                            // uncompressed size               4 bytes
+      . pack32le($gzLength)                     // compressed size                 4 bytes
+      . pack32le($dataLength)                   // uncompressed size               4 bytes
       . pack16le(strlen($filePath))             // file name length                2 bytes
       . pack16le(strlen($zip64Ext))             // extra field length              2 bytes
       . $filePath                               // file name                       (variable size)
       . $zip64Ext;                              // extra field                     (variable size)
   }
 
+  private function addDataDescriptor($dataLength, $gzLength, $dataCRC32) {
+    if ($this->zip64) {
+      $length = 20;
+      $packedGzLength = pack64le($gzLength);
+      $packedDataLength = pack64le($dataLength);
+    } else {
+      $length = 12;
+      $packedGzLength = pack32le($gzLength->getLoBytes());
+      $packedDataLength = pack32le($dataLength->getLoBytes());
+     }
+
+    $this->write(''
+        . pack32le($dataCRC32)  // crc-32                          4 bytes
+        . $packedGzLength       // compressed size                 4/8 bytes (depending on zip64 enabled)
+        . $packedDataLength     // uncompressed size               4/8 bytes (depending on zip64 enabled)
+        .'');
+    return $length;
+  }
+
   private function buildZip64EndOfCentralDirectoryRecord($cdRecLength) {
+    $versionToExtract = $this->getVersionToExtract(False);
     $cdRecCount = sizeof($this->cdRec);
 
     return ''
@@ -325,7 +448,7 @@ class ZipStreamer {
         . pack64le(44)                                   // size of zip64 end of central directory
                                                          // record                                     8 bytes
         . pack16le(self::ATTR_MADE_BY_VERSION)           //version made by                             2 bytes
-        . pack16le(self::ATTR_VERSION_TO_EXTRACT)        //version needed to extract                   2 bytes
+        . pack16le($versionToExtract)                    // version needed to extract                  2 bytes
         . pack32le(0)                                    // number of this disk                        4 bytes
         . pack32le(0)                                    // number of the disk with the start of the
                                                          // central directory                          4 bytes
@@ -341,7 +464,7 @@ class ZipStreamer {
   }
 
   private function buildZip64EndOfCentralDirectoryLocator($cdRecLength) {
-    $zip64RecStart = Count64::construct($this->offset)->add($cdRecLength);
+    $zip64RecStart = Count64::construct($this->offset, !$this->zip64)->add($cdRecLength);
 
         return ''
         . pack32le(self::ZIP64_END_OF_CENTRAL_DIR_LOCATOR) // zip64 end of central dir locator signature  4 bytes  (0x07064b50)
@@ -353,47 +476,71 @@ class ZipStreamer {
   }
 
   private function buildCentralDirectoryHeader($filePath, $timestamp, $gpFlags,
-      $gzMethod, $dataLength, $gzLength, $dataCRC32, $extFileAttr) {
+      $gzMethod, $dataLength, $gzLength, $dataCRC32, $extFileAttr, $isDir) {
+    $versionToExtract = $this->getVersionToExtract($isDir);
     $dosTime = self::getDosTime($timestamp);
-    $zip64Ext = $this->buildZip64ExtendedInformationField($dataLength, $gzLength);
+    if ($this->zip64) {
+      $zip64Ext = $this->buildZip64ExtendedInformationField($dataLength, $gzLength);
+      $dataLength = -1;
+      $gzLength = -1;
+      $diskNo = -1;
+      $offset = -1;
+    } else {
+      $zip64Ext = '';
+      $dataLength = $dataLength->getLoBytes();
+      $gzLength = $gzLength->getLoBytes();
+      $diskNo = 0;
+      $offset = $this->offset->getLoBytes();
+    }
 
     return ''
       . pack32le(self::ZIP_CENTRAL_FILE_HEADER)  //central file header signature   4 bytes  (0x02014b50)
       . pack16le(self::ATTR_MADE_BY_VERSION)     //version made by                 2 bytes
-      . pack16le(self::ATTR_VERSION_TO_EXTRACT)  //version needed to extract       2 bytes
+      . pack16le($versionToExtract)              // version needed to extract      2 bytes
       . pack16le($gpFlags)                       //general purpose bit flag        2 bytes
       . pack16le($gzMethod)                      //compression method              2 bytes
       . pack32le($dosTime)                       //last mod file time              2 bytes
                                                  //last mod file date              2 bytes
       . pack32le($dataCRC32)                     //crc-32                          4 bytes
-      . pack32le(-1)                             //compressed size                 4 bytes
-      . pack32le(-1)                             //uncompressed size               4 bytes
+      . pack32le($gzLength)                      //compressed size                 4 bytes
+      . pack32le($dataLength)                    //uncompressed size               4 bytes
       . pack16le(strlen($filePath))              //file name length                2 bytes
       . pack16le(strlen($zip64Ext))              //extra field length              2 bytes
       . pack16le(0)                              //file comment length             2 bytes
-      . pack16le(-1)                             //disk number start               2 bytes
+      . pack16le($diskNo)                        //disk number start               2 bytes
       . pack16le(0)                              //internal file attributes        2 bytes
       . pack32le($extFileAttr)                   //external file attributes        4 bytes
-      . pack32le(-1)                             //relative offset of local header 4 bytes
+      . pack32le($offset)                        //relative offset of local header 4 bytes
       . $filePath                                //file name                       (variable size)
       . $zip64Ext                                //extra field                     (variable size)
       //TODO: implement?
       . '';                                      //file comment                    (variable size)
   }
 
-  private function buildEndOfCentralDirectoryRecord() {
+  private function buildEndOfCentralDirectoryRecord($cdRecLength) {
+    if ($this->zip64) {
+      $diskNumber = -1;
+      $cdRecCount = -1;
+      $cdRecLength = -1;
+      $offset = -1;
+    } else {
+      $diskNumber = 0;
+      $cdRecCount = sizeof($this->cdRec);
+      $offset = $this->offset->getLoBytes();
+    }
+    //throw new \Exception(sprintf("zip64 %d diskno %d", $this->zip64, $diskNumber));
 
     return ''
       . pack32le(self::ZIP_END_OF_CENTRAL_DIRECTORY) // end of central dir signature    4 bytes  (0x06064b50)
-      . pack16le(-1)                                 // number of this disk             2 bytes
-      . pack16le(-1)                                 // number of the disk with the
+      . pack16le($diskNumber)                        // number of this disk             2 bytes
+      . pack16le($diskNumber)                        // number of the disk with the
                                                      // start of the central directory  2 bytes
-      . pack16le(-1)                                 // total number of entries in the
+      . pack16le($cdRecCount)                        // total number of entries in the
                                                      // central directory on this disk  2 bytes
-      . pack16le(-1)                                 // total number of entries in the
+      . pack16le($cdRecCount)                        // total number of entries in the
                                                      // central directory               2 bytes
-      . pack32le(-1)                                 // size of the central directory   4 bytes
-      . pack32le(-1)                                 // offset of start of central
+      . pack32le($cdRecLength)                       // size of the central directory   4 bytes
+      . pack32le($offset)                            // offset of start of central
                                                      // directory with respect to the
                                                      // starting disk number            4 bytes
       . pack16le(0)                                  // .ZIP file comment length        2 bytes
@@ -413,7 +560,7 @@ class ZipStreamer {
    * @param int $timestamp
    * @return 2-byte encoded DOS Date
    */
-  private static function getDosTime($timestamp = 0) {
+  public static function getDosTime($timestamp = 0) {
     $timestamp = (int) $timestamp;
     $oldTZ = @date_default_timezone_get();
     date_default_timezone_set('UTC');
@@ -429,13 +576,15 @@ class ZipStreamer {
 
 abstract class ExtFileAttr {
 
-  # ZIP external file attributes layout
-  # TTTTsstrwxrwxrwx0000000000ADVSHR
-  # ^^^^____________________________ UNIX file type
-  #     ^^^_________________________ UNIX setuid, setgid, sticky
-  #        ^^^^^^^^^________________ UNIX permissions
-  #                 ^^^^^^^^________ "lower-middle byte" (TODO: what is this?)
-  #                         ^^^^^^^^ DOS attributes (reserved, reserved, archived, directory, volume, system, hidden, read-only
+  /*
+    ZIP external file attributes layout
+    TTTTsstrwxrwxrwx0000000000ADVSHR
+    ^^^^____________________________ UNIX file type
+        ^^^_________________________ UNIX setuid, setgid, sticky
+           ^^^^^^^^^________________ UNIX permissions
+                    ^^^^^^^^________ "lower-middle byte" (TODO: what is this?)
+                            ^^^^^^^^ DOS attributes (reserved, reserved, archived, directory, volume, system, hidden, read-only
+  */
 
   public static function getExtFileAttr($attr) {
     return $attr;
@@ -444,7 +593,7 @@ abstract class ExtFileAttr {
 
 class UNIX extends ExtFileAttr {
 
-  # Octal
+  // Octal
   const S_IFIFO = 0010000; /* named pipe (fifo) */
   const S_IFCHR = 0020000; /* character special */
   const S_IFDIR = 0040000; /* directory */
@@ -474,6 +623,88 @@ class UNIX extends ExtFileAttr {
   }
 }
 
+abstract class DeflateStream {
+  static public function create($level) {
+    if (COMPR::NONE === $level) {
+      return new DeflateStoreStream($level);
+    } else {
+      return new DeflatePeclStream($level);
+    }
+  }
+  protected function __construct($level) {}
+
+  abstract public function update($data);
+  abstract public function finish();
+}
+
+class DeflatePeclStream extends DeflateStream {
+  private $peclDeflateStream;
+
+  const PECL1_DEFLATE_STREAM_CLASS = '\HttpDeflateStream';
+  const PECL2_DEFLATE_STREAM_CLASS = '\http\encoding\Stream\Deflate';
+
+  protected function __construct($level) {
+    $class = self::PECL1_DEFLATE_STREAM_CLASS;
+    if (!class_exists($class)) {
+      $class = self::PECL2_DEFLATE_STREAM_CLASS;
+    }
+    if (!class_exists($class)) {
+      new \Exception('unable to instantiate PECL deflate stream (requires pecl_http >= 0.10)');
+    }
+
+    $deflateFlags = constant($class . '::TYPE_RAW');
+    switch ($level) {
+      case COMPR::NORMAL:
+        $deflateFlags |= constant($class . '::LEVEL_DEF');
+        break;
+      case COMPR::MAXIMUM:
+        $deflateFlags |= constant($class . '::LEVEL_MAX');
+        break;
+      case COMPR::SUPERFAST:
+        $deflateFlags |= constant($class . '::LEVEL_MIN');
+        break;
+    }
+    $this->peclDeflateStream = new $class($deflateFlags);
+  }
+
+  public function update($data) {
+    return $this->peclDeflateStream->update($data);
+  }
+
+  public function finish() {
+    return $this->peclDeflateStream->finish();
+  }
+}
+
+class DeflateStoreStream extends DeflateStream {
+  const BLOCK_HEADER_NORMAL = 0x00;
+  const BLOCK_HEADER_FINAL = 0x01;
+  const BLOCK_HEADER_ERROR = 0x03;
+
+  const MAX_UNCOMPR_BLOCK_SIZE = 0xffff;
+
+  public function update($data) {
+    $result = '';
+    for ($pos = 0, $len = strlen($data); $pos < $len; $pos += self::MAX_UNCOMPR_BLOCK_SIZE) {
+      $result .= $this->write_block(self::BLOCK_HEADER_NORMAL, substr($data, $pos, self::MAX_UNCOMPR_BLOCK_SIZE));
+    }
+    return $result;
+  }
+
+  public function finish() {
+    return $this->write_block(self::BLOCK_HEADER_FINAL, '');
+  }
+
+  private function write_block($header, $data) {
+    return ''
+        . pack8($header)                    // block header                     3 bits, null padding = 1 byte
+        . pack16le(strlen($data))           // block data length                2 bytes
+        . pack16le(0xffff ^ strlen($data))  // complement of block data size    2 bytes
+        . $data                             // data
+        . '';
+      }
+}
+
 class DOS extends ExtFileAttr {
 
   const READ_ONLY = 0x1;
@@ -487,12 +718,16 @@ class DOS extends ExtFileAttr {
 }
 
 class GPFLAGS {
+  const NONE = 0x0000; // no flags set
+  const COMP1 = 0x0002; // compression flag 1 (compression settings, see APPNOTE for details)
+  const COMP2 = 0x0004; // compression flag 2 (compression settings, see APPNOTE for details)
   const ADD = 0x0008; // ADD flag (sizes and crc32 are append in data descriptor)
   const EFS = 0x0800; // EFS flag (UTF-8 encoded filename and/or comment)
-}
 
-class GZMETHOD {
-  const STORE = 0x0000; //  0 - The file is stored (no compression)
-  const DEFLATE = 0x0008; //  8 - The file is Deflated
+  // compression settings for deflate/deflate64
+  const DEFL_NORM = 0x0000; // normal compression (COMP1 and COMP2 not set)
+  const DEFL_MAX = COMP1; // maximum compression
+  const DEFL_FAST = COMP2; // fast compression
+  const DEFL_SFAST = 0x0006; // superfast compression (COMP1 and COMP2 set)
 }
 
